
#
# C# exporter for Gaphor. Tested with Gaphor 0.11.2.
# Copyright 2006-2007 Juri Pakaste <juri@iki.fi>
# This software is in the public domain.
#
# TODO (at least):
# - Extensive testing
# - Better namespace handling
# 
# To use types not visible by default, add tags with the name using,
# value packagename.
#
# To create a delegate, create a class with one operation and give the
# class the tag delegate with value true. The delegate in the code
# will have the name of the class in the diagram and the parameter
# list and the return value of the operation. The name of the
# operation will be discarded.
#
# To create an event, define in the containing class a tag with the name
# event and set the value to be the name of the property you want to be an 
# event.
#
# The replacements file is a file containing entries looking like this:
#
# uml id (something like DCE:4ADFA748-E9A6-11DB-888B-00112F116AE6)
# %%% comment
# replacement type (one of the types defined in ReplacementReader.TYPES)
# text
# %%
#
# It allows you to replace declarations, bodies, getters and setters of
# the various generated elements. Every entry must be terminated by
# %%. Lines starting with %%% are considered comments. Empty lines are
# ignored before the id and type.
#
# The uml id line can be a single id or a composite consisting of the
# parent's id and the element's id, separated by space. This is useful
# for methods implementing classes get from interfaces.
#
# To debug the C# generation, set the environment variable
# GAPHOR_CSEXPORT_DEBUG
#


import os
import pprint
import sys
import traceback

import pdb

import gtk

from gaphor import UML
from gaphor.core import _, inject, action, build_action_group
from gaphor.interfaces import IService, IActionProvider
from zope import interface, component


import ReplacementReader

INDENT_LEVEL = 4

export_folder = None
export_file = None
replacement_file = None

class CSharpExportAction(object):
    title = "Export the diagram to a C# file"
    ext = ".cs"

    interface.implements(IService, IActionProvider)

    element_factory = inject('element_factory')
    gui_manager = inject('gui_manager')

    menu_xml = """
      <ui>
        <menubar action="mainwindow">
          <menu action="file">
            <menu action="file-export">
              <menuitem action="file-export-csharp" />
            </menu>
          </menu>
        </menubar>
      </ui>"""

    def __init__(self):
        self.action_group = build_action_group(self)
        self._latest_dir = os.getcwd()

    def init(self, app):
        pass

    def shutdown(self):
        pass

    def show_export_dialog(self):
        dialog = gtk.Dialog(title="Export Model to C#",
                            buttons=(gtk.STOCK_CANCEL,gtk.RESPONSE_CANCEL,
                                     gtk.STOCK_EXECUTE,gtk.RESPONSE_OK))
        vbox = gtk.VBox()
        vbox.set_spacing(6)

        def create_alignment():
            a = gtk.Alignment()
            a.set_padding(6, 0, 12, 0)
            return a

        WIDTH = 20

        # target selection
        target_frame = gtk.Frame("<b>Select export target</b>")
        target_frame.get_label_widget().set_use_markup(True)
        target_frame.set_shadow_type(gtk.SHADOW_NONE)
        target_alignment = create_alignment()

        target_vbox = gtk.VBox()

        target_chooser = gtk.FileChooserButton("Export Folder")
        target_chooser.set_action(gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER)
        target_chooser.set_width_chars(WIDTH)
        if export_folder is not None and export_file is not None:
            target_chooser.set_filename(os.path.join(export_folder, export_file))
        
        target_entry = gtk.Entry()
        if export_file is not None:
            target_entry.set_text(export_file)
        target_entry.set_width_chars(WIDTH)

        target_vbox.pack_start(target_chooser)
        target_vbox.pack_start(target_entry)
        target_vbox.set_spacing(6)

        target_alignment.add(target_vbox)
        target_frame.add(target_alignment)

        # replacements selection
        replacement_frame = gtk.Frame("<b>Select replacements file</b>")
        replacement_frame.get_label_widget().set_use_markup(True)
        replacement_frame.set_shadow_type(gtk.SHADOW_NONE)
        replacement_alignment = create_alignment()

        replacement_vbox = gtk.VBox()

        replacement_chooser = gtk.FileChooserButton("Replacement File")
        replacement_chooser.set_action(gtk.FILE_CHOOSER_ACTION_OPEN)
        replacement_chooser.set_width_chars(WIDTH)
        if replacement_file is not None:
            replacement_chooser.set_filename(replacement_file)
        
        replacement_vbox.pack_start(replacement_chooser)
        replacement_vbox.set_spacing(6)

        replacement_alignment.add(replacement_vbox)
        replacement_frame.add(replacement_alignment)

        vbox.add(target_frame)
        vbox.add(replacement_frame)
        vbox.show_all()

        dialog.vbox.pack_start(vbox)

        response = dialog.run()
        if response == gtk.RESPONSE_OK:
            target = os.path.join(target_chooser.get_filename(), target_entry.get_text())
            global export_folder, export_file, replacement_file
            export_folder, export_file = os.path.split(target)
            replacements = replacement_chooser.get_filename()
            replacement_file = replacements
        else:
            target = None
            replacements = None
        dialog.destroy()
        return (target, replacements)

    def read_export_filename(self):
        chooser = gtk.FileChooserDialog(
            title="Select directory and file where to generate C#",
            action=gtk.FILE_CHOOSER_ACTION_SAVE,
            buttons=(gtk.STOCK_CANCEL,gtk.RESPONSE_CANCEL,
                     gtk.STOCK_OPEN,gtk.RESPONSE_OK))
        rootelem = root(self.element_factory)
        chooser.set_current_name("%s.cs" % rootelem.name)
        response = chooser.run()
        if response == gtk.RESPONSE_OK:
            filename = chooser.get_filename()
        elif response == gtk.RESPONSE_CANCEL:
            filename = None
        else:
            log.error("strange response: %s" % response)
            filename = None
        chooser.destroy()
        return filename
        
    def read_replacement_filename(self):
        chooser = gtk.FileChooserDialog(
            title="Select a replacements file",
            action=gtk.FILE_CHOOSER_ACTION_OPEN,
            buttons=(gtk.STOCK_CANCEL,gtk.RESPONSE_CANCEL,
                     gtk.STOCK_OPEN,gtk.RESPONSE_OK))
        response = chooser.run()
        if response == gtk.RESPONSE_OK:
            filename = chooser.get_filename()
        elif response == gtk.RESPONSE_CANCEL:
            filename = None
        chooser.destroy()
        return filename

    @action(name='file-export-csharp', label=_('Export to C#'),
            tooltip=_('Export model to a C# file'))
    def execute(self):
        """gaphor's plugin main callback"""
        filename, replacements = self.show_export_dialog()
        if filename is not None:
            self.export(filename, replacements)
        return


        filename = self.read_export_filename()
        if filename is not None:
            replacements = self.read_replacement_filename()
            self.export(filename, replacements)

    def export(self, filename, replacements):
        r = root(self.element_factory)
        if os.getenv("GAPHOR_CSEXPORT_DEBUG") is not None:
            pdb.runcall(csconverter.run, filename, replacements, r, self.element_factory)
        else:
            csconverter.run(filename, replacements, r, self.element_factory)

def root(factory):
    """return the root package element from the gaphor elements factory"""
    roots = [elmt for elmt in factory._elements.values()
             if isinstance(elmt, UML.Package) and elmt.namespace is None]
    #assert len(roots) == 1, roots
    return roots[0]

def path(elmt):
    """return the path to this gaphor element from the root package"""
    if elmt.namespace is None:
        return [elmt.name]
    return path(elmt.namespace) + [elmt.name]

class VarnameString(str):
    def capitalize_first(self):
        if len(self) > 0:
            return self[0].upper() + self[1:]
        return ""

    def uncapitalize_first(self):
        if len(self) > 0:
            return self[0].lower() + self[1:]
        return ""

class ReplacementList(list):
    def find_type(self, type, default=None):
        t = [r for r in self if r.type == type]
        if len(t):
            return t[-1]
        return default

class ReplacementDict(dict):
    def getr(self, k):
        if self.has_key(k):
            return self[k]
        return ReplacementList((ReplacementReader.get_empty_replacement(),))

class Gaphor2CSharp:
    def run(self, filename, replacements, root, element_factory):
        self._uids = {}
        self.element_factory = element_factory
        replacements = self.read_replacements(replacements)
        self._file = open(filename, "w")
        try:
            r = self.visit(root)
            self.write_using(r.get_packages(), self._file)
            r.output(0, self._file, replacements)
        finally:
            self._file.close()

    def read_replacements(self, filename):
        rd = ReplacementDict()
        if filename:
            for r in ReplacementReader.read_replacements(open(filename, "r")):
                l = rd.get(r.id, ReplacementList())
                l.append(r)
                log.debug("added replacement %s" % r)
                rd[r.id] = l
        return rd

    def write_using(self, packages, file):
        for p in packages:
            file.write("using %s;\n" % p)
        file.write("\n")

    def add_tags_to_cselement(self, elmt, csel):
        if elmt.taggedValue is not None:
            for tv in elmt.taggedValue:
                t, v = tv.value.split("=", 1)
                csel.add_tag(t, v)

    def visit(self, elmt):
        if self._uids.has_key(elmt.id):
            return self._uids[elmt.id]
        attrname = "visit_%s" % elmt.__class__.__name__.lower()
        if hasattr(self, attrname):
            log.debug("calling %s for %s" % (attrname, elmt.name))
            return getattr(self, attrname)(elmt)
        else:
            log.debug("method %s not defined, skipping %s" % (attrname, elmt.name))
            return None

    def visit_package(self, elmt):
        ns = CSNamespace(elmt.id, elmt.name)
        self._uids[elmt.id] = ns
        self.add_tags_to_cselement(elmt, ns)
        members = [m for m in elmt.member]
        members.sort(key=lambda m: m.name)
        for sub in members:
            s = self.visit(sub)
            if isinstance(s, CSClass):
                ns.add_class(s)
            elif isinstance(s, CSNamespace):
                ns.add_namespace(s)
            elif isinstance(s, CSInterface):
                ns.add_interface(s)
            elif isinstance(s, CSDelegate):
                ns.add_delegate(s)
            else:
                log.error("Don't know what to do with package %s's sub element %s" % (ns, s))
        return ns
    
    def visit_class(self, elmt):
        super = None
        interfaces = []
        struct = False
        abstract = False
        visibility = "public"
        new = False

        for impl in elmt.implementation:
            for contract in impl.contract:
                i = self.visit(contract)
                if isinstance(i, CSInterface):
                    interfaces.append(i)
        delegate = False
        if elmt.taggedValue is not None:
            for tv in elmt.taggedValue:
                t, v = tv.value.split("=", 1)
                if t == "delegate" and v == "true":
                    delegate = True
        if delegate:
            c = CSDelegate(elmt.id, elmt.name, visibility)
            self._uids[elmt.id] = c
            if len(elmt.ownedOperation) != 1:
                log.error("A Delegate should have exactly one operation, %s had %d" % (elmt.name, len(elmt.ownedOperation)))
                return None
            else:
                log.debug("Found one operation for delegate %s" % elmt.name)
                m = self.visit(elmt.ownedOperation[0])
                c.set_operation(m)
                log.debug("returning %s" % c)
                return c
        else:
            c = CSClass(elmt.id, elmt.name, super, interfaces, struct, 
                        abstract, visibility, new)
            self.add_tags_to_cselement(elmt, c)
            self._uids[elmt.id] = c

            # There's no UI to add tags to properties and looks like inserting
            # them by hand to the file doesn't work either. To mark properties
            # as events, we have to define event=propertyname tags in the class
            # and fish them out here.
            events = []
            for t, v in [tv.value.split("=", 1) for tv in (elmt.taggedValue or [])]:
                log.debug("tags of class %s: %s = %s" % (c.name, t, v))
                if t == "event":
                    events.append(v)

            ownedattrs = [attr for attr in elmt.ownedAttribute]
            ownedattrs.sort(key=lambda a: a.name)
            for attr in ownedattrs:
                p = self.visit(attr)
                if p is not None:
                    c.add_property(p)
                    if p.name in events:
                        p.event = True
                    
            ownedops = [op for op in elmt.ownedOperation]
            ownedops.sort(key=lambda o: o.name)
            for op in ownedops:
                m = self.visit(op)
                if m is not None:
                    c.add_method(m)
            return c

    def visit_property(self, elmt):
        access = "public"
        ptype = ""
        event = False
        const = False
        readonly = False
        static = False
        volatile = False
        getter = True
        setter = True
        value = None

        extra_tags = []

        lowervalue = elmt.lowerValue.value if elmt.lowerValue is not None else None
        uppervalue = elmt.upperValue.value if elmt.upperValue is not None else None
        typevalue = elmt.typeValue.value if elmt.typeValue is not None else None
        visibility = elmt.visibility
        namespace = elmt.namespace
        taggedvalue = elmt.taggedValue
        etype = elmt.type.name if elmt.type is not None else None

        access = visibility
        ns = self.visit(namespace)
        if typevalue is not None:
            ptype = typevalue
        elif etype is not None:
            ptype = etype
        else:
            log.error("type of property %s in %s can't be determined" % (elmt.name, ns))
            return None
        
        if uppervalue is not None and uppervalue != "1":
            ptype = "List<%s>" % ptype
            extra_tags.append(("using", "System.Collections.Generic"))
        else:
            log.debug("upperValue didn't indicate multiplicity: %s" % uppervalue)

        for t, v in [tv.split("=", 1) for tv in (taggedvalue or [])]:
            log.debug("tag for %s: %s = %s" % (elmt.name, t, v))
            if t == 'event' and v == 'true':
                event = True

        p = CSProperty(elmt.id, elmt.name, ptype, access, event, const, 
                       readonly, static, volatile, getter, setter, value)
        self._uids[elmt.id] = p
        self.add_tags_to_cselement(elmt, p)
        for tn, tv in extra_tags:
            p.add_tag(tn, tv)
        return p

    def visit_operation(self, elmt):
        access = "public"
        rtype = None

        ns = self.visit(elmt.namespace)
        #print "namespace of operation %s: %s" % (elmt.name, ns)
        try:
            rtype = elmt.returnResult.items[0].typeValue.value
        except:
            log.error("Couldn't determine the return type of method %s" % elmt.name)
        if rtype is None:
            log.error("Couldn't determine the return type of method %s" % elmt.name)
            rtype = "void"
        m = CSMethod(
            elmt.id, elmt.name, access, rtype, 
            [(e.typeValue.value, e.name) for e in elmt.formalParameter])
        self._uids[elmt.id] = m
        self.add_tags_to_cselement(elmt, m)
        return m

#     def visit_association(self, elmt):
#         log.debug("association %s, skipping" % elmt)

#     def visit_interface(self, elmt):
#         log.debug("interface %s, skipping" % elmt)

#    def visit_diagram(self, elmt):
#        return self.visit_package(elmt)

    def visit_interface(self, elmt):
        
        vis = elmt.visibility
        i = CSInterface(elmt.id, elmt.name, elmt.visibility)
        self._uids[elmt.id] = i
        self.add_tags_to_cselement(elmt, i)

        # There's no UI to add tags to properties and looks like inserting
        # them by hand to the file doesn't work either. To mark properties
        # as events, we have to define event=propertyname tags in the class
        # and fish them out here.
        events = []
        for t, v in [tv.value.split("=", 1) for tv in (elmt.taggedValue or [])]:
            log.debug("tags of interface %s: %s = %s" % (i.name, t, v))
            if t == "event":
                events.append(v)

        log.debug("looking for interface %s's supers" % elmt.name)
        for gen in elmt.generalization:
            log.debug("gen: %s" % gen)
            for target in gen.target:
                log.debug("target: %s" % target)
                super = self.visit(target)
                log.debug("super: %s" % super)
                i.add_super(super)
        
        ownedattrs = [attr for attr in elmt.ownedAttribute]
        ownedattrs.sort(key=lambda a: a.name)
        for attr in ownedattrs:
            p = self.visit(attr)
            i.add_property(p)
            if p.name in events:
                p.event = True
        ownedops = [op for op in elmt.ownedOperation]
        ownedops.sort(key=lambda a: a.name)
        for op in ownedops:
            i.add_method(self.visit(op))
        return i

class CSElement(object):
    def __init__(self, id, name):
        self._tags = []
        self._id = id
        self._name = VarnameString(name)

    def get_name(self):
        return self._name
    name = property(get_name, None, None, None)

    def get_id(self):
        return self._id
    id = property(get_id, None, None, None)

    def p(self, level, fo, text):
        for line in text.split("\n"):
            fo.write(level * INDENT_LEVEL * " ")
            fo.write(line)
            fo.write("\n")

    def add_tag(self, tag, value):
        self._tags.append((tag, value))

    def tag_values(self, tag):
        vals = [v for (t, v) in self._tags if t == tag]
        return vals

    def get_child_packages(self):
        return set()

    def get_packages(self):
        return self.get_child_packages().union(self.tag_values("using"))
    packages = property(get_packages, None, None, "Packages used by this element")

    def __str__(self):
        return "<%s %s>" % (self.__class__.__name__, self._name)

class CSInterface(CSElement):
    __slots__ = ("_visibility", "_properties", "_methods", "_name", "_supers",
                 "interface", "_id")

    def __init__(self, id, name, visibility):
        CSElement.__init__(self, id, name)
        self._visibility = visibility
        self._methods = []
        self._properties = []
        self._supers = []
        self.interface = True

    def add_method(self, m): self._methods.append(m)
    def add_property(self, p): self._properties.append(p)
    def add_super(self, s): self._supers.append(s)
    
    def get_methods(self):
        return tuple(self._methods)
    
    def get_properties(self):
        return tuple(self._properties)

    def get_all_methods(self):
        sm = []
        for s in self._supers:
            sm.extend(s.methods)
        return tuple(self._methods + sm)

    def get_all_properties(self):
        sp = []
        for s in self._supers:
            sp.extend(s.properties)
        return tuple(self._properties + sp)

    methods = property(get_methods, None, None, "The methods")
    properties = property(get_properties, None, None, "The properties")

    all_methods = property(get_all_methods, None, None, 
                           "Methods from the hierarchy")
    all_properties = property(get_all_properties, None, None, 
                              "Properties from the hierarchy")

    def output(self, level, fo, replacements):
        r = replacements.getr(self.id)[-1]
        if r.type == ReplacementReader.REPLACE_WHOLE:
            self.p(level, r.text)
        else:
            if r.type == ReplacementReader.REPLACE_DECLARATION:
                decl = [r.text]
            else:
                decl = [self._visibility, "interface", self._name]
                if self._supers:
                    decl.append(':')
                    decl.append(", ".join(s.name for s in self._supers))
            decl.append('{')
            self.p(level, fo, " ".join(decl))
            self.p(level+1, fo, "/* uml %s */" % self.id)
            if r.type == ReplacementReader.REPLACE_BODY:
                self.p(level + 1, r.text)
            else:
                for p in self._properties:
                    p.output(level + 1, fo, replacements, self)
                for m in self._methods:
                    m.output(level + 1, fo, replacements, self)
            if r.type == ReplacementReader.APPEND_BODY:
                self.p(level + 1, r.text)
            self.p(level, fo, "}")

class CSClass(CSElement):
    __slots__ = ("_properties", "_methods", "_classes", 
                 "_name", "_super", "_interfaces",
                 "_struct", "_abstract", "_visibility", "_new", "interface",
                 "_id")

    def __init__(self, id, name, super, interfaces, struct, abstract, visibility, new):
        CSElement.__init__(self, id, name)
        self._super = super
        self._interfaces = interfaces

        self._struct = struct
        self._abstract = abstract
        self._visibility = visibility
        self._new = new
        
        self._properties = []
        self._methods = []
        self._classes = []
        if struct and super:
            log.warn("A superclass (%s) defined for a struct (%s)", super, name)
        if struct and abstract:
            log.warn("A struct (%s) declared as abstract", name)
        self.interface = False

    def add_property(self, prop): self._properties.append(prop)
    def add_method(self, meth): self._methods.append(meth)
    def add_class(self, klass): self._classes.append(klass)

    def get_child_packages(self):
        pkgs = set()
        for c in self._classes:
            pkgs.update(c.get_packages())
        for m in self._methods:
            pkgs.update(m.get_packages())
        for p in self._properties:
            pkgs.update(p.get_packages())
        return pkgs

    def output(self, level, fo, replacements):
        rep = replacements.getr(self.id)
        r = rep.find_type(ReplacementReader.REPLACE_WHOLE)
        if r:
            self.p(level, fo, r.text)
        else:
            r = rep.find_type(ReplacementReader.REPLACE_DECLARATION)
            if r:
                decl = [r.text]
            else:
                decl = [self._visibility]
                if self._abstract:
                    decl.append("abstract")
                if self._struct:
                    decl.append("struct")
                else:
                    decl.append("class")
                decl.append(self._name)
                if self._super or self._interfaces:
                    decl.append(":")
                    extends = []
                    if self._super:
                        extends.append(self._super)
                    if self._interfaces:
                        extends.extend([i.name for i in self._interfaces])
                    decl.append(", ".join(extends))
            decl.append("{")
            self.p(level, fo, " ".join(decl))
            self.p(level+1, fo, "/* uml %s */" % self.id)
            r = rep.find_type(ReplacementReader.REPLACE_BODY)
            if r:
                self.p(level+1, fo, r.text)
            else:
                for c in self._classes:
                    c.output(level + 1, fo, replacements)
                for c in self._properties:
                    c.output(level + 1, fo, replacements, self)
                for c in self._methods:
                    c.output(level + 1, fo, replacements, self)
                for i in self._interfaces:
                    self.p(level + 1, fo, 
                           "/* Implementation for interface %s */" % i.name)
                    for m in i.all_methods:
                        m.output(level + 1, fo, replacements, self)
                    for p in i.all_properties:
                        p.output(level + 1, fo, replacements, self)
                r = rep.find_type(ReplacementReader.APPEND_BODY)
                if r:
                    self.p(level+1, fo, r.text)
            self.p(level, fo, "}")

    def __str__(self):
        return "<%s %s>" % (self.__class__.__name__, self._name)

class CSNamespace(CSElement):
    __slots__ = ("_name", "_classes", "_namespaces", "_interfaces")

    def __init__(self, id, name):
        CSElement.__init__(self, id, name)
        self._classes = []
        self._namespaces = []
        self._interfaces = []
        self._delegates = []

    def add_class(self, klass): self._classes.append(klass)
    def add_namespace(self, namespace): self._namespaces.append(namespace)
    def add_interface(self, interface): self._interfaces.append(interface)
    def add_delegate(self, delegate): self._delegates.append(delegate)

    def get_child_packages(self):
        pkgs = set()
        for c in self._classes:
            pkgs.update(c.get_packages())
        return pkgs

    def output(self, level, fo, replacements):
        r = replacements.getr(self.id)[-1]
        if r.type == ReplacementReader.REPLACE_WHOLE:
            self.p(level, fo, r.text)
        else:
            if r.type == ReplacementReader.REPLACE_DECLARATION:
                self.p(level, fo, r.text)
            else:
                self.p(level, fo, "namespace %s" % self._name)
            self.p(level, fo, "{")
            self.p(level+1, fo, "/* uml %s */" % self.id)
            if r.text == ReplacementReader.REPLACE_BODY:
                self.p(level+1, fo, r.text)
            else:
                for c in self._namespaces:
                    c.output(level + 1, fo, replacements)
                for c in self._delegates:
                    c.output(level + 1, fo, replacements)
                for c in self._interfaces:
                    c.output(level + 1, fo, replacements)
                for c in self._classes:
                    c.output(level + 1, fo, replacements)
                if r.type == ReplacementReader.APPEND_BODY:
                    self.p(level+1, fo, r.text)
            self.p(level, fo, "}")

    def __str__(self):
        return "<%s %s>" % (self.__class__.__name__, self._name)

class CSProperty(CSElement):
    def __init__(self, id, name, type, access, event, const, readonly, static, volatile, getter, setter, value):
        CSElement.__init__(self, id, name)
        self._type = type
        self._access = access
        self._const = const
        self._readonly = readonly
        self._event = event
        self._static = static
        self._volatile = volatile
        self._getter = getter
        self._setter = setter
        self._value = value

    def get_event(self): return self._event
    def set_event(self, event): self._event = event
    event = property(get_event, set_event, None, "Is the property an event")

    def output(self, level, fo, replacements, parent):
        if parent.interface:
            wholeid = self.id
        else:
            wholeid = "%s %s" % (parent.id, self.id)
        wrep = replacements.getr(wholeid)
        rep = replacements.getr(self.id)
        wr = wrep.find_type(ReplacementReader.REPLACE_WHOLE)
        r = rep.find_type(ReplacementReader.REPLACE_WHOLE)

        if wr is not None:
            self.p(level, fo, wr.text)
        elif r is not None:
            self.p(level, fo, r.text)

        else:
            decl = []
            # we always make vars private, props get the access level declared
            # in the uml.
            if self._event:
                if not parent.interface:
                    decl.append(self._access)
                decl.append("event")
                decl.append(self._type)
                decl.append(self.name)
                decl.append(";")
            elif not parent.interface:
                decl.append("private")
                if self._const: decl.append("const")
                if self._readonly: decl.append("readonly")
                if self._static: decl.append("static")
                if self._volatile: decl.append("volatile")
                decl.append(self._type)
                varname = self.name.uncapitalize_first()
                decl.append(varname)
                decl.append(";")
            gs = []
            if not self._event:
                wr = wrep.find_type(ReplacementReader.REPLACE_GETTER)
                r = rep.find_type(ReplacementReader.REPLACE_GETTER)
                if wr is not None:
                    gs.append(wr.text)
                elif r is not None:
                    gs.append(r.text)
                else:
                    if self._getter == "private":
                        gs.append("private")
                    if self._getter:
                        gs.append("get")
                        if parent.interface:
                            gs.append(";")
                        else:
                            gs.append("{ return %s; }" % varname)
                
                wr = wrep.find_type(ReplacementReader.REPLACE_SETTER)
                r = rep.find_type(ReplacementReader.REPLACE_SETTER)
                if wr is not None:
                    gs.append(wr.text)
                elif r is not None:
                    gs.append(r.text)
                else:
                    if self._setter == "private":
                        gs.append("private")
                    if self._setter:
                        gs.append("set")
                        if parent.interface:
                            gs.append(";")
                        else:
                            gs.append("{ %s = value; }" % varname)
                if gs:
                    propname = self.name.capitalize_first()
                    if not parent.interface:
                        decl.append(self._access)

                    decl.append(self._type)
                    decl.append(propname)
                    decl.append("{")
                    decl.extend(gs)
                    decl.append("}")
            decl.append("/* uml %s */" % self.id)
            self.p(level, fo, " ".join(decl))

    def __str__(self):
        return "<%s %s>" % (self.__class__.__name__, self._name)

class CSMethod(CSElement):
    def __init__(self, id, name, access, rtype, params):
        CSElement.__init__(self, id, name)
        self._access = access
        self._rtype = rtype
        self._params = params

    def get_rtype(self):
        return self._rtype
    rtype = property(get_rtype, None, None, "The return type")

    def get_params(self):
        return tuple(self._params)
    params = property(get_params, None, None, "The parameters")

    def output(self, level, fo, replacements, parent):
        log.debug("outputting method %s: access: %s parent: %s (interface: %s)" % (self.name, self._access, parent.name, parent.interface))
        if parent.interface:
            wholeid = self.id
        else:
            wholeid = "%s %s" % (parent.id, self.id)
        log.debug("method wholeid: %s" % wholeid)
        wrep = replacements.getr(wholeid)
        rep = replacements.getr(self.id)
        log.debug("method %s: replacements with self.id: %s" % (self.name, [str(r) for r in rep]))
        log.debug("method %s: replacements with whole id: %s" % (
                self.name, [str(r) for r in wrep]))
        wr = wrep.find_type(ReplacementReader.REPLACE_WHOLE)
        r = rep.find_type(ReplacementReader.REPLACE_WHOLE)

        if wr is not None:
            self.p(level, fo, wr.text)
        elif r is not None:
            self.p(level, fo, r.text)
        else:
            wr = wrep.find_type(ReplacementReader.REPLACE_DECLARATION)
            r = rep.find_type(ReplacementReader.REPLACE_DECLARATION)
            if wr is not None:
                dec = wr.text
            elif r is not None:
                dec = r.text
            else:
                decl = []
                if not parent.interface:
                    decl.append(self._access)
                decl.extend([self._rtype, self._name])
                plist = []
                for tp in self._params:
                    plist.append(" ".join(tp))
                if parent.interface:
                    term = "; /* uml %s */" % self.id
                else:
                    term = ""
                dec = "%s(%s)%s" % (" ".join(decl), ", ".join(plist), term)
            self.p(level, fo, dec)
            if not parent.interface:
                self.p(level, fo, "{")
                self.p(level+1, fo, "/* uml %s %s */" % (parent.id, self.id))
                rtypes = (ReplacementReader.REPLACE_BODY, ReplacementReader.APPEND_BODY)
                for wr in wrep:
                    if wr.type in rtypes:
                        self.p(level+1, fo, wr.text)
                        break
                for r in rep:
                    if r.type in rtypes:
                        self.p(level+1, fo, r.text)
                        break
                self.p(level, fo, "}")

    def __str__(self):
        return "<%s %s>" % (self.__class__.__name__, self._name)

class CSDelegate(CSElement):
    def __init__(self, id, name, visibility):
        CSElement.__init__(self, id, name)
        self._visibility = visibility

    def set_operation(self, op):
        self._dname = op.name
        self._rtype = op.rtype
        self._params = op.params

    def output(self, level, fo, replacements, parent=None):
        log.debug("outputting CSDelegate %s" % self.name)
        decl = [self._visibility, "delegate", self._rtype, self._name]
        plist = []
        r = replacements.getr(self.id)[-1]
        if r.type in (ReplacementReader.REPLACE_WHOLE, ReplacementReader.REPLACE_DECLARATION):
            self.p(level, fo, r.text)
        else:
            for tp in self._params:
                plist.append(" ".join(tp))
            self.p(level, fo, "%s(%s);%s" % (
                    " ".join(decl), 
                    ", ".join(plist),
                    "/* uml %s */" % self.id))

csconverter = Gaphor2CSharp()

