import textwrap
import unittest

TYPES = {
    'NOTHING': 0,
    'REPLACE_BODY': 1,
    'REPLACE_DECLARATION': 2,
    'REPLACE_WHOLE': 3,
    'APPEND_BODY': 4,
    'REPLACE_GETTER': 5,
    'REPLACE_SETTER': 6
}

# copy the type names into the module scope
for t,v in TYPES.iteritems():
    locals()[t] = v

SEPARATOR = '%%'

empty_replacement = None

def get_empty_replacement():
    global empty_replacement
    if empty_replacement is None:
        empty_replacement = Replacement("", NOTHING, "")
    return empty_replacement

class Replacement(object):
    __slots__ = ('_id', '_type', '_text')
    def __init__(self, id = None, type = None, text = ""):
        self._id = id
        self._type = type
        self._text = text

    def get_id(self): return self._id
    def set_id(self, id):
        assert self._id is None, "can't set property multiple times"
        self._id = id
    id = property(get_id, set_id, None, "Id of the replacement")

    def get_type(self): return self._type
    def set_type(self, type):
        assert self._type is None, "can't set property multiple times"
        assert type in TYPES.values(), "type must be one of %s" % ", ".join(TYPES.values())
        self._type = type
    type = property(get_type, set_type, None, "Type of the replacement")

    def get_text(self): return self._text
    def set_text(self, text):
        self._text = text
    text = property(get_text, set_text, None, "Text of the replacement")

    def __str__(self):
        return "<%s id: %s type: %s>" % (self.__class__.__name__, self.id, self.type)

STATE_READ_ID = 1
STATE_READ_TYPE = 2
STATE_READ_TEXT = 3

def read_replacements(file):
    rep = Replacement()
    text = []
    state = STATE_READ_ID
    for line in file:
        if line != '':
            sline = line.strip()
            if sline == SEPARATOR:
                if len(text):
                    rep.text = textwrap.dedent("\n".join(text))
                else:
                    rep.text = ""
                if rep.id is not None and rep.type is not None:
                    yield rep
                rep = Replacement()
                text = []
                state = STATE_READ_ID
            elif line.startswith("%%%"):
                continue
            else:
                if state == STATE_READ_ID:
                    if sline:
                        rep.id = sline
                        state = STATE_READ_TYPE
                elif state == STATE_READ_TYPE:
                    if sline:
                        rep.type = TYPES[sline]
                        state = STATE_READ_TEXT
                else:
                    text.append(line.rstrip())

class TestReadReplacements(unittest.TestCase):
    def _test_replacements(self, lines, count, tests):
        g = read_replacements(lines)
        cc = count
        tc = 0
        while cc > 0:
            try:
                r = g.next()
                if tc < len(tests):
                    t = tests.pop(tc)
                    tc += 1
                    t(r)
            except StopIteration:
                self.fail("Expected %d replacements, got only %d" % (count, count - cc))
            cc -= 1
        extra = 0
        for line in g:
            extra += 1
        if extra:
            self.fail("Expected %d replacements, got %d" % (count, extra))

    def test_empty(self):
        self._test_replacements([], 0, [])
    
    def test_empty_sep(self):
        def t(r):
            self.assertEqual(None, r.id)
            self.assertEqual(None, r.type)
            self.assertEqual("", r.text)
        self._test_replacements([SEPARATOR], 1, [t])

    def test_one_id(self):
        id = 'DCE:foo-bar'
        def t(r):
            self.assertEqual(id, r.id)
            self.assertEqual(None, r.type)
            self.assertEqual("", r.text)
        self._test_replacements([id, SEPARATOR], 1, [t])

    def test_one_id_and_type(self):
        id = 'DCE:foo-bar'
        type = 'REPLACE_BODY'
        def t(r):
            self.assertEqual(id, r.id)
            self.assertEqual(TYPES[type], r.type)
            self.assertEqual("", r.text)
        self._test_replacements([id, type, SEPARATOR], 1, [t])

    def test_one_full(self):
        id = 'DCE:foo-bar'
        type = 'REPLACE_BODY'
        text = "line1"
        def t(r):
            self.assertEqual(id, r.id)
            self.assertEqual(TYPES[type], r.type)
            self.assertEqual(text, r.text)
        self._test_replacements([id, type, text, SEPARATOR], 1, [t])

    def test_one_full_multiline(self):
        id = 'DCE:foo-bar'
        type = 'REPLACE_BODY'
        text = "line1\nline2"
        def t(r):
            self.assertEqual(id, r.id)
            self.assertEqual(TYPES[type], r.type)
            self.assertEqual(text, r.text)
        self._test_replacements([id, type] + text.split() + [SEPARATOR], 1, [t])

    def test_two_full_multiline(self):
        id1 = 'DCE:foo-bar'
        type1 = 'REPLACE_BODY'
        text1 = "line1\nline2"
        def t1(r):
            self.assertEqual(id1, r.id)
            self.assertEqual(TYPES[type1], r.type)
            self.assertEqual(text1, r.text)

        id2 = 'DCE:foo-bar-2'
        type2 = 'REPLACE_BODY'
        text2 = "line1\nline2\nline3"
        def t2(r):
            self.assertEqual(id2, r.id)
            self.assertEqual(TYPES[type2], r.type)
            self.assertEqual(text2, r.text)

        self._test_replacements([id1, type1] + text1.split() + [SEPARATOR, id2, type2] + 
                                text2.split() + [SEPARATOR], 
                                2, [t1, t2])

if __name__ == "__main__":
    unittest.main()
